class ReplayFile{
	int64 lastModified;
	string path;
	
	string mapUID;
	int pb;
	
	ReplayFile(const string &in path){
		this.path= path;
		lastModified = IO::FileModifiedTime(path);
	}	
	
	void ExtractInfo(){
		IO::File f(path);
		f.Open(IO::FileMode::Read);
		f.SetPos(0xe0); // might also first search for string GetLocalLogin()
		string line = f.ReadLine();
		f.Close();
		mapUID = SubStrDelim(line, "<map uid=\"", "\"");
		pb = Text::ParseInt(SubStrDelim(line, "<times best=\"", "\""));
		if(mapUID == "" || pb <= 0){
			error("Problem reading replay file "+path);
		}
	}
	
	int opCmp(ReplayFile@ other) {
		return this.lastModified > other.lastModified ? 1 : (this.lastModified < other.lastModified ? -1 : 0);
	}
}

class TrainingData{
	MapInfo@[] mapInfos;
	int[] pbs;
	
	TrainingData(){
	}
	
	TrainingData(Json::Value v){
		if(v.GetType() != Json::Type::Array) {
			error("TrainingData(Json::Value v): v is no array");
			return;
		}
		
		for(uint i = 0; i < v.Length; i++){
			string uid = v[i]["mapUID"];
			if(storage is null) {
				error("TrainingData(Json::Value v): storage null");
				return;
			}
			MapInfo@ mapInfo = storage.GetMapInfo(uid);
			if(mapInfo is null) {
				error("TrainingData(Json::Value v): map not in stroage.");
				return;
			}
			mapInfos.InsertLast(mapInfo);
			pbs.InsertLast(v[i]["pb"]);
		}
	}
		
	Json::Value@ ToJson() const{
		auto a = Json::Array();
		for(uint i = 0; i < mapInfos.Length; i++){
			auto v = Json::Object();
			v["mapUID"] = mapInfos[i].mapUID;
			v["pb"] = pbs[i];
			a.Add(v);
		}
		return a;
	}
	
	void RefreshFromReplayFiles(){
		Log("====== Refreshing Training Data from files");
		mapInfos = {};
		pbs = {};
		
		auto fileList = IO::IndexFolder(IO::FromUserGameFolder("Replays\\Autosaves"),false);
		array<ReplayFile@> replays;
		for(uint i = 0; i < fileList.Length; i++){
			replays.InsertLast(ReplayFile(fileList[i]));
		}
		replays.SortDesc();
		
		
		uint i = 0;
		while(mapInfos.Length < nbTrainingData){
			if(i >= replays.Length){
				error("Not enough valid replay files.");
				return;
			}
			auto replay = replays[i++];
			Log("Analyzing replay "+replay.path, false);
			replay.ExtractInfo();
			Log(", UID = "+replay.mapUID);
			auto mapInfo = storage.GetMapInfo(replay.mapUID);
			if(mapInfo is null){
				Log(" Map not in storage, calling init", false);
				@mapInfo = MapInfo(replay.mapUID);
				if(mapInfo.Init() <= 2) //we skip if init failed, or LB not closed
					continue;
				Log(", and adding map to storage");
				storage.AddMapInfo(@mapInfo);
			} 
			else 
				Log(" Map found in storage");
			
			if(replay.pb > mapInfo.silverTime) {
				Log(" Skipping map: pb="+replay.pb+">"+mapInfo.silverTime+"=silver"); 
				continue;
			}
			mapInfos.InsertLast(mapInfo);
			pbs.InsertLast(replay.pb);
		}
	}
}

class MapInfo{
	string mapUID;
	string seasonUID;
	
	int authorTime;
	int goldTime;
	int silverTime;
	int bronzeTime;
	
	string[] top100;
	int[] top100Times;
	
	MapInfo(const string&in mapUID){
		this.mapUID = mapUID;
	}
	
	MapInfo(Json::Value a){
		if(a.GetType() != Json::Type::Object) {
			error("MapInfo: called with bad Json::Value"); 
			return;
		}
		
		mapUID = a["mapUID"];
		seasonUID = a["seasonUID"];
		authorTime = a["authorTime"];
		goldTime = a["goldTime"];
		silverTime = a["silverTime"];
		bronzeTime = a["bronzeTime"];
		auto jsonTop100 = a["top100"];
		auto jsonTop100Times = a["top100Times"];

		if(jsonTop100.Length < 100 || jsonTop100Times.Length < 100) {
			error("MapInfo: less than 100 top100 in Json::value"); 
			return;
		}
		for(uint i = 0; i < 100; i++){
			top100.InsertLast(jsonTop100[i]);
			top100Times.InsertLast(jsonTop100Times[i]);
		}
	}
	
	MapInfo(CGameCtnChallengeInfo@ mapInfo){
		mapUID = mapInfo.MapUid;
		authorTime = mapInfo.TMObjective_AuthorTime;
		goldTime = mapInfo.TMObjective_GoldTime;
		silverTime = mapInfo.TMObjective_SilverTime;
		bronzeTime = mapInfo.TMObjective_BronzeTime;
	}
	
	bool IsInit(){
		return (seasonUID != "" && top100.Length == 100 && top100Times.Length == 100);
	}
	
	int Init(){	//codes: neg: fail, positive: pass
		if(IsInit()) {
			Log("   (Init) "+mapUID+": map already init"); 
			return 1;
		}
		
		int uidI = storage.totdList.IndexOf(mapUID);
		if(uidI == -1) {
			Log("   (Init) "+mapUID+": not found in totdlist"); 
			return -1;
		}

		string base = storage.totdList.SubStr(uidI,250);
		string[] infos = base.SubStr(0, base.IndexOf("\n")).Split(",");
		seasonUID = infos[1];
		int startTimeStamp = Text::ParseInt(infos[2]);

		auto resp = LiveGet("/api/token/leaderboard/group/"+seasonUID+"/map/"+mapUID+"/top?length=100&onlyWorld=true&offset=0");
		int i = resp.IndexOf("\"accountId\"");
		while(i > 0){
			resp = resp.SubStr(i+13);
			top100.InsertLast(resp.SubStr(0, accountLength));
			top100Times.InsertLast(Text::ParseInt( SubStrDelim(resp, "\"score\":", "}")));
			i = resp.IndexOf("\"accountId\"");
		}
		if(top100.Length < 100 || top100Times.Length < 100) {
			Log("   (Init) "+mapUID+": problem in top100, length="+top100.Length);
			return -2;
		}
		
		if(authorTime == 0){
			auto mapInfoStr = LiveGet("/api/token/map/"+mapUID);
			authorTime = Text::ParseInt(SubStrDelim(mapInfoStr,"\"authorTime\":", ","));
			goldTime = Text::ParseInt(SubStrDelim(mapInfoStr,"\"goldTime\":", ","));
			silverTime = Text::ParseInt(SubStrDelim(mapInfoStr,"\"silverTime\":", ","));
			bronzeTime = Text::ParseInt(SubStrDelim(mapInfoStr,"\"bronzeTime\":", ","));

			Log("   "+mapUID+": parsed "+ authorTime + ", "+ goldTime + ", "+ silverTime + ", "+ bronzeTime);
			if(authorTime <= 0 || goldTime <= 0 || silverTime <= 0 || bronzeTime <= 0){ 
				Log("   (Init) "+mapUID+": problem with parsing");
				return -3;
			}
		}
		
		if(startTimeStamp > Time::Stamp - 60*60*24){
			Log("   (Init) "+mapUID+": init ok, but LB not closed"); 
			return 2;
		}
	
		return 3;
	}
	
	double correlateAgainst(MapInfo@ other){
		double corr = 0;
		for(int i = 0; i < 100; i++){
			int j = other.top100.Find(top100[i]);
			if(j != -1){
				corr += double(1.0)/double((1 + (i*i)/100.0) * (1 + (j*j)/100.0));
			}
		}
		return corr;
	}
	
	Json::Value@ ToJson() const{
		auto a = Json::Object();
		a["mapUID"] = Json::Value(mapUID);
		a["authorTime"] = authorTime;
		a["goldTime"] = goldTime;
		a["silverTime"] = silverTime;
		a["bronzeTime"] = bronzeTime;
		a["seasonUID"] = Json::Value(seasonUID);
		a["top100"] = Json::Value(top100);
		a["top100Times"] = Json::Value(top100Times);
		return a;
	}
}

