[Setting category="General" name="Window visiblility hotkey"]
VirtualKey windowVisibleKey = VirtualKey(VirtualKey::F8);

[Setting category="Debug" name="log" hidden]
string logStr = "";
[Setting category="Debug" name="debug" hidden]
bool debug = false;
[SettingsTab name="Debug"]
void RenderSettings()
{
	debug = UI::Checkbox("Enable Debug mode", debug);

	if(debug){
		UI::BeginTable("##main", 4,UI::TableFlags::SizingFixedFit);
		UI::TableNextColumn();
		if(UI::Button("Compute Goal Time")) 
			startnew(SetCurrentGoalTime);
		UI::TableNextColumn();
		if(UI::Button("Refresh Training Data")) 
			startnew(RefreshTrainingData);
		UI::TableNextColumn();
		if(UI::Button("Refresh Totd List")) 
			startnew(RefreshTotdList);
		UI::TableNextColumn();
		if(UI::Button("Clear Log")) logStr = "";
		UI::EndTable();
	
		UI::InputTextMultiline("Log", logStr, vec2 ( 700 , 200 ), UI::InputTextFlags::ReadOnly);
  	}
}

//constants
const uint accountLength = 36;
const uint seasonUIDLength = 36;
const int maxLogLength = 20000;
const int nbTrainingData = 30;

const double improvement = 0.2; //20% improvement
const double cutoff = 1.75; //bellow that we're not correlated
const double cutoff2 = 1; //above that + cutoff we don't add dummy correlation

void Log(const string&in str, bool newline = true){
	if(!debug)
		return;
	while(logStr.Length >= maxLogLength)
		logStr = logStr.SubStr(logStr.IndexOf("\n")+1);
	logStr += str;
	if(newline){
		logStr += "\n";
	}
}

// substring extraction
string SubStrDelim(const string&in str, const string&in begin, const string&in end){
	string tmp = str.SubStr(str.IndexOf(begin) + begin.Length);
	return tmp.SubStr(0, tmp.IndexOf(end));
}

UI::Font@ font;
UI::Texture@ goalImage;
UI::Texture@ goalImageGreen;
CTrackMania@ app;
Storage@ storage;
int currentGoalTime = -1; //-1 = unset, 0 = no goal time
int currentPB = -1;
void Main() {
	@app = cast<CTrackMania>(GetApp());
	@storage = Storage();
	storage.LoadAll();

	@font = UI::LoadFont("DroidSans.ttf", 16, -1, -1, true, true, true);
	@goalImage = UI::LoadTexture("goal.png");
	@goalImageGreen = UI::LoadTexture("goal-green.png");

	bool firstRefresh = storage.trainingData is null;
	if(firstRefresh){
		UI::ShowNotification("Improve: First Load",
		"The plugin is loading the training data for the first time. This might take a minute, you'll be notified when it's done.",
		vec4(0.27f, 0.27f, 0, 1),
		4000);
	}

	if(storage.latestTimeStamp < Time::Stamp - 60*60*24 - 60)
		storage.RefreshTotdList();
	

	// refresh training data if not present, or every 9-10 days
	if(storage.trainingData is null || storage.trainingData.mapInfos.Length < nbTrainingData 
		|| storage.LastTrainingDataRefresh < Time::Stamp - 60*60*24*9)
		storage.RefreshTrainingData();

	if(firstRefresh && storage.trainingData !is null && storage.trainingData.mapInfos.Length == nbTrainingData ){
		UI::ShowNotification("Improve: First Load Done",
		"Training data loading complete, goal times should now generate automatically.",
		vec4(0, 0.27f ,0 , 1),
		4000);
	}

	while(true){
		if(storage.latestTimeStamp < Time::Stamp - 60*60*24 - 60)
			storage.RefreshTotdList();

		if(app.RootMap !is null && app.RootMap.MapInfo.MapUid != ""){
			if(currentGoalTime == -1){
				currentGoalTime = storage.GetGoalTime(app.RootMap.MapInfo.MapUid);
				if(currentGoalTime == -1){ //auto compute goal if not present
					SetCurrentGoalTime();
				}
			}
			
			//set PB
			auto network = cast<CTrackManiaNetwork>(app.Network);
			auto userMgr = network.ClientManiaAppPlayground.UserMgr;
			MwId userId;
			if (userMgr.Users.Length > 0) {
				userId = userMgr.Users[0].Id;
			} else {
				userId.Value = uint(-1);
			}
			auto scoreMgr = network.ClientManiaAppPlayground.ScoreMgr;
			currentPB = scoreMgr.Map_GetRecord_v2(userId, app.RootMap.MapInfo.MapUid, "PersonalBest", "", "TimeAttack", "");
		} else {
			currentGoalTime = -1;
			currentPB = -1;
		}

		sleep(500);
	}
}

void Render(){
	auto map = app.RootMap;

	if(!UI::IsGameUIVisible() || map is null ||  map.MapInfo.MapUid == "" || app.Editor !is null || 
		!windowVisible || currentGoalTime == 0) 
		return;
	
	UI::SetNextWindowPos(0, 40, UI::Cond::FirstUseEver);
	
	int windowFlags = UI::WindowFlags::NoTitleBar 
						| UI::WindowFlags::NoCollapse 
						| UI::WindowFlags::AlwaysAutoResize 
						| UI::WindowFlags::NoDocking;
	UI::PushFont(font);
	UI::Begin("GoalTime", windowFlags);
	UI::BeginTable("##main", 2);
    UI::TableNextRow();

	UI::Texture@ goal = @goalImage;
	string prefix = "";
	if(currentGoalTime <= 0){
		prefix = "\\$888";
	} else if (currentPB > 0 && currentPB <= currentGoalTime){
		@goal = @goalImageGreen;
		prefix = "\\$0b0";
	}
	UI::TableNextColumn();
	UI::Image(goal, vec2(16,16));
	UI::TableNextColumn();
	if(currentGoalTime <= 0) 
		UI::Text(prefix + "unset");
	else 
		UI::Text(prefix + Time::Format(currentGoalTime));	
		
    UI::EndTable();
	UI::End();
	UI::PopFont();
}
	
// util class to store correlation and score
class CorrScore{
	double corr;
	double score;
	CorrScore(double corr, double score){
		this.corr = corr;
		this.score = score;
	}

	// score
	int opCmp(CorrScore@ other) {
		return this.score > other.score ? 1 : (this.score < other.score ? -1 : 0);
	}

	// corr
	/*int opCmp(CorrScore@ other) {
		return this.corr > other.corr ? 1 : (this.corr < other.corr ? -1 : 0);
	}*/
}

int ComputeGoalTime(MapInfo@ mapInfo){
	Log("====== Computing Correlation for "+ mapInfo.mapUID);	
	
	array<CorrScore@> csArray;
	auto mapInfos = @storage.trainingData.mapInfos;
	auto pbs = @storage.trainingData.pbs;
	
	double totalCorr = 0;
	for(uint i =0; i < mapInfos.Length; i++){ 
		if(mapInfo.mapUID == mapInfos[i].mapUID) //skipping self map
			continue;
		double corr = mapInfo.correlateAgainst(mapInfos[i]) - cutoff; 
		if(corr > 0) {
			csArray.InsertLast(CorrScore(corr, double(pbs[i]) / double(mapInfos[i].top100Times[0])));
			totalCorr += corr;
			if(debug)
				Log(" ("+corr+","+mapInfos[i].top100Times[0]+","+pbs[i]+", "+double(pbs[i]) / double(mapInfos[i].top100Times[0])+")",false);
			
		}
	}
	Log("");

	// no corr found, return gold
	if(csArray.Length == 0){
		Log("No corr found, returning gold");
		return mapInfo.goldTime;
	}

	yield();
	
	if(debug){
		csArray.SortDesc();
		for(uint i =0; i < csArray.Length; i++){ 
			Log(" ("+csArray[i].corr+", "+csArray[i].score+") ",false);
		}
		Log("");
	}
	
	//adding dummy correlation with gold if not at cutoff2
	if(totalCorr < cutoff2){
		double dummyCorr = cutoff2 - totalCorr;
		csArray.InsertLast(CorrScore(dummyCorr, double(mapInfo.goldTime) / double(mapInfo.top100Times[0])));
		totalCorr += dummyCorr;
		Log("Adding dummy corr " + dummyCorr);
	} 
	else{ // we remove the lowest scores
		csArray.SortDesc();
		double low = Math::Min(totalCorr/4.0,totalCorr - cutoff2); // removes the bottom quarter, or what remains to get totalCorr = cutoff2
		Log("Removing "+low +"/"+totalCorr+" of low times");
		totalCorr -= low;
		while(low > 0){
			if(csArray[0].corr <= low){
				low -= csArray[0].corr;
				csArray.RemoveAt(0);
			} 
			else{
				csArray[0].corr -= low;
				low = 0;
			}
		}
	}

	if(debug){
		for(uint i =0; i < csArray.Length; i++){ 
			Log(" ("+csArray[i].corr+", "+csArray[i].score+")", false);
		}
		Log("");
	}
	
	//we compute the coeff that should match PB/top1
	double c = 0;
	for(uint i =0; i < csArray.Length; i++){ 
		c += csArray[i].corr * csArray[i].score;
	}
	c /= totalCorr;
	
	double predictedTime = double(mapInfo.top100Times[0])*c;

	//we push the pred time towards the top 1 with improvement factor
	int goalTime = int((1-improvement) * predictedTime + improvement * mapInfo.top100Times[0]); 
	
	if(mapInfo.goldTime <= goalTime){
		Log("Goal time "+goalTime+" worse than gold, returning gold");
		return mapInfo.goldTime;
	}
	
	//glue goalTime to author time.
	if( mapInfo.authorTime - 50 < goalTime && goalTime < mapInfo.authorTime +200 ){
		goalTime = mapInfo.authorTime;
		Log("Gluing goal time "+goalTime+" to author");
	}
	
	Log("Found "+csArray.Length+" maps. Predtime = "+int(predictedTime)+", c="+c);
	return goalTime;
}

int SetGoalTime(CGameCtnChallengeInfo@ gameMapInfo){
	auto mapInfo = storage.GetMapInfo(gameMapInfo.MapUid);
	int state = 0;
	if(mapInfo is null){
		@mapInfo = MapInfo(gameMapInfo);
		state = mapInfo.Init();
		if(state <= 0) // fail
			return 0; 
		if(state == 3) // init ok and LB closed
			storage.AddMapInfo(mapInfo); 
	}
	
	int goalTime = ComputeGoalTime(mapInfo);
	if(state != 2) // we don't store goaltime if LB not closed
		storage.SetGoalTime(gameMapInfo.MapUid, goalTime);

	return goalTime;
}

void SetCurrentGoalTime(){
	if(app.RootMap is null || app.RootMap.MapInfo is null){
		error("Attempting to set goal time without being in a map.");
		return;
	}
	currentGoalTime = SetGoalTime(app.RootMap.MapInfo);
}

void RefreshTrainingData(){
	storage.RefreshTrainingData();
}

void RefreshTotdList(){
	storage.RefreshTotdList();
}

bool held = false;
bool windowVisible = true;
bool debugVisible = true;
void OnKeyPress(bool down, VirtualKey key)
{
	if(key == windowVisibleKey && !held)
	{
		windowVisible = !windowVisible;
	}
	held = down;
}

bool liveInit = false;
string LiveGet(const string&in url){
	if(!liveInit){
		NadeoServices::AddAudience("NadeoLiveServices");
		while(!NadeoServices::IsAuthenticated("NadeoLiveServices")) yield();
		liveInit=true;
	}
	
	auto req = NadeoServices::Get("NadeoLiveServices",NadeoServices::BaseURLLive()+ url);
	req.Start();
	while(!req.Finished()) yield();
	return req.String();
}